/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.maven.ban;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;
import org.apache.maven.artifact.versioning.InvalidVersionSpecificationException;
import org.apache.maven.artifact.versioning.VersionRange;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.plugin.MojoFailureException;
import org.codehaus.plexus.util.xml.Xpp3Dom;
import org.eclipse.aether.artifact.Artifact;
import org.eclipse.aether.artifact.DefaultArtifact;
import org.eclipse.aether.impl.ArtifactResolver;
import org.eclipse.aether.impl.VersionRangeResolver;
import org.eclipse.aether.resolution.ArtifactRequest;
import org.eclipse.aether.resolution.ArtifactResolutionException;
import org.eclipse.aether.resolution.ArtifactResult;
import org.eclipse.aether.resolution.VersionRangeRequest;
import org.eclipse.aether.resolution.VersionRangeResolutionException;
import org.eclipse.aether.resolution.VersionRangeResult;
import org.eclipse.aether.version.Version;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.inteligr8.maven.model.ArtifactFilter;

public abstract class AbstractBanConfiguration implements BanConfiguration {
    
    private final Logger logger = LoggerFactory.getLogger(this.getClass());
    private final Pattern artifactPattern = Pattern.compile("^([^:]*):([^:]*)(:([^:]+))?$");
    private final Pattern notRegexPattern = Pattern.compile("^[A-Za-z0-9_\\-\\.]*$");
    
    protected final List<ArtifactFilter> includeArtifacts = new LinkedList<>();
    protected final List<ArtifactFilter> excludeArtifacts = new LinkedList<>();

    private final ArtifactResolver artifactResolver;
    private final VersionRangeResolver versionRangeResolver;
    private final MavenSession session;
    
    public AbstractBanConfiguration(MavenSession session, ArtifactResolver artifactResolver, VersionRangeResolver versionRangeResolver) {
        this.session = session;
        this.artifactResolver = artifactResolver;
        this.versionRangeResolver = versionRangeResolver;
    }
    
    public void init(Xpp3Dom rootDom) throws IOException, MojoFailureException {
        if (rootDom == null)
            return;

        Xpp3Dom importDom = rootDom.getChild("import");
        if (importDom != null)
            this.processImports(importDom);
        this.processIncludesExcludes(rootDom);
    }
    
    private void processImports(Xpp3Dom importDom) throws IOException, MojoFailureException {
        for (Xpp3Dom child : importDom.getChildren()) {
            BanConfigurationDownloader downloader = null;
            if (child.getName().equals("file")) {
                File file = new File(StringUtils.trimToNull(child.getValue()));
                downloader = new BanConfigurationDownloader(this.session, this.artifactResolver, this.versionRangeResolver, file);
            } else if (child.getName().equals("url")) {
                String url = StringUtils.trimToNull(child.getValue());
                downloader = new BanConfigurationDownloader(this.session, this.artifactResolver, this.versionRangeResolver, url);
            } else if (child.getName().equals("artifact")) {
                Artifact artifact = new DefaultArtifact(child.getValue());
                if (!"xml".equals(artifact.getExtension()))
                    artifact = new DefaultArtifact(artifact.getGroupId(), artifact.getArtifactId(), artifact.getClassifier(), "xml", artifact.getVersion());

                Version latestLocalVersion = this.findLatestLocalVersion(artifact, child.getValue());
                Artifact latestArtifact = this.findLatestArtifact(artifact, child.getValue());
                if (latestArtifact == null && latestLocalVersion != null) {
                    this.logger.debug("A latest version was found locally, but could not resolve the artifact; trying to resolve the artifact with the specific version: {}: {}", latestLocalVersion, child.getValue());
                    artifact = artifact.setVersion(latestLocalVersion.toString());
                    latestArtifact = this.findLatestArtifact(artifact, child.getValue());
                }
                
                if (artifact != null) {
                    File file = artifact.getFile();
                    downloader = new BanConfigurationDownloader(this.session, this.artifactResolver, this.versionRangeResolver, file);
                }
            } else {
                this.logger.debug("Unrecognized configuration element ignored: {}: {}", child.getName(), child.getValue());
            }
            
            if (downloader != null) {
                this.includeArtifacts.addAll(downloader.getIncludeArtifacts());
                this.excludeArtifacts.addAll(downloader.getExcludeArtifacts());
            }
        }
    }
    
    private Version findLatestLocalVersion(Artifact artifact, String logId) {
        this.logger.trace("Inspecting the local repository to select the version to import: {}", logId);
        VersionRangeRequest vrrequest = new VersionRangeRequest(artifact, this.session.getCurrentProject().getRemoteProjectRepositories(), null);
        try {
            VersionRangeResult vrresult = this.versionRangeResolver.resolveVersionRange(this.session.getRepositorySession(), vrrequest);
            if (vrresult.getVersions().isEmpty()) {
                this.logger.info("The artifact version range could not be resolved locally; trying remote: {}", logId);
                return null;
            } else {
                this.logger.debug("The artifact version discovered locally: {}; trying remote: {}", vrresult.getHighestVersion(), logId);
                return vrresult.getHighestVersion();
            }
        } catch (VersionRangeResolutionException vrre) {
            this.logger.error("The artifact version range could not be resolved; skipping: {}", logId);
            return null;
        }
    }
    
    private Artifact findLatestArtifact(Artifact artifact, String logId) {
        this.logger.trace("Inspecting the local/remote repositories to select the artifact to import: {}", logId);
        ArtifactRequest arequest = new ArtifactRequest(artifact, this.session.getCurrentProject().getRemoteProjectRepositories(), null);
        try {
            ArtifactResult aresult = this.artifactResolver.resolveArtifact(this.session.getRepositorySession(), arequest);
            this.logger.debug("This artifact version discovered remotely: {}: {}", aresult.getArtifact().getVersion(), logId);
            return aresult.getArtifact();
        } catch (ArtifactResolutionException are) {
            this.logger.warn("The artifact could not be resolved; skipping: {}", artifact);
            return null;
        }
    }
    
    private void processIncludesExcludes(Xpp3Dom rootDom) {
        Xpp3Dom includesDom = rootDom.getChild("includes");
        if (includesDom != null)
            this.includeArtifacts.addAll(this.parseArtifacts(includesDom));
        
        Xpp3Dom excludesDom = rootDom.getChild("excludes");
        if (excludesDom != null)
            this.excludeArtifacts.addAll(this.parseArtifacts(excludesDom));
        
        this.logger.debug("Include artifacts: {}", this.includeArtifacts);
        this.logger.debug("Exclude artifacts: {}", this.excludeArtifacts);
    }
    
    public List<ArtifactFilter> getIncludeArtifacts() {
        return this.includeArtifacts;
    }
    
    public List<ArtifactFilter> getExcludeArtifacts() {
        return this.excludeArtifacts;
    }
    
    private List<ArtifactFilter> parseArtifacts(Xpp3Dom artifactsDom) {
        List<ArtifactFilter> filters = new LinkedList<>();
        
        for (Xpp3Dom artifactDom : artifactsDom.getChildren("artifact")) {
            ArtifactFilter filter = new ArtifactFilter();
            String versionSpec = null;
            if (artifactDom.getChildCount() == 0) {
                Matcher matcher = this.artifactPattern.matcher(artifactDom.getValue());
                if (!matcher.matches()) {
                    this.logger.warn("The artifact format '{}' does not match the expected regular expression: {}; ignoring ...", artifactDom.getValue(), this.artifactPattern.pattern());
                    continue;
                }
                
                if (this.notRegexPattern.matcher(matcher.group(1)).matches()) {
                    filter.setGroupId(StringUtils.trimToNull(matcher.group(1)));
                } else {
                    filter.setGroupIdRegex(StringUtils.trimToNull(matcher.group(1)));
                }
                
                if (this.notRegexPattern.matcher(matcher.group(2)).matches()) {
                    filter.setArtifactId(StringUtils.trimToNull(matcher.group(2)));
                } else {
                    filter.setArtifactIdRegex(StringUtils.trimToNull(matcher.group(2)));
                }
                
                versionSpec = StringUtils.trimToNull(matcher.group(4));
            } else {
                filter.setGroupId(this.getChildValue(artifactDom, "groupId"));
                filter.setGroupIdRegex(this.getChildValue(artifactDom, "groupIdRegex"));
                filter.setArtifactId(this.getChildValue(artifactDom, "artifactId"));
                filter.setArtifactIdRegex(this.getChildValue(artifactDom, "artifactIdRegex"));

                versionSpec = this.getChildValue(artifactDom, "version");
            }

            if (versionSpec != null) {
                try {
                    VersionRange versionRange = VersionRange.createFromVersionSpec(versionSpec);
                    filter.setVersionRange(versionRange);
                } catch (InvalidVersionSpecificationException ivse) {
                    this.logger.warn("The artifact '{}' has an invalid version specification; the artifact element will be ignored: {}", ivse.getMessage());
                    continue;
                }
            }

            filters.add(filter);
        }
        
        return filters;
    }
    
    private String getChildValue(Xpp3Dom dom, String child) {
        Xpp3Dom childDom = dom.getChild(child);
        return childDom == null ? null : StringUtils.trimToNull(childDom.getValue());
    }
    
}
