/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.maven.ban;

import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.maven.artifact.versioning.DefaultArtifactVersion;
import org.apache.maven.artifact.versioning.VersionRange;
import org.eclipse.aether.artifact.Artifact;
import org.eclipse.aether.graph.DependencyFilter;
import org.eclipse.aether.graph.DependencyNode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.inteligr8.maven.model.ArtifactFilter;

public class BanDependencyFilter implements DependencyFilter {
    
    private final Logger logger = LoggerFactory.getLogger(this.getClass());
    
    private final List<ArtifactFilter> includeArtifacts;
    private final List<ArtifactFilter> excludeArtifacts;
    private boolean failFast = false;
    
    public BanDependencyFilter(List<ArtifactFilter> includeArtifacts, List<ArtifactFilter> excludeArtifacts) {
        this.includeArtifacts = includeArtifacts;
        this.excludeArtifacts = excludeArtifacts;
    }
    
    public void setFailFast(boolean failFast) {
        this.failFast = failFast;
    }
    
    @Override
    public boolean accept(DependencyNode node, List<DependencyNode> parents) {
        this.logger.debug("Evaluating dependency '{}'", node);
        Artifact depArtifact = node.getArtifact();
        boolean ban = false;
        
        for (ArtifactFilter afilter : this.includeArtifacts) {
            if (this.matches(afilter.getGroupId(), afilter.getGroupIdRegex(), depArtifact.getGroupId()) &&
                    this.matches(afilter.getArtifactId(), afilter.getArtifactIdRegex(), depArtifact.getArtifactId()) &&
                    this.withinRange(afilter.getVersionRange(), depArtifact.getVersion())) {
                this.logger.debug("The dependency '{}' matches the ban inclusion filter", depArtifact);
                ban = true;
                break;
            }
        }
        
        if (!ban)
            return false;
        
        for (ArtifactFilter afilter : this.excludeArtifacts) {
            if (this.matches(afilter.getGroupId(), afilter.getGroupIdRegex(), depArtifact.getGroupId()) &&
                    this.matches(afilter.getArtifactId(), afilter.getArtifactIdRegex(), depArtifact.getArtifactId()) &&
                    this.withinRange(afilter.getVersionRange(), depArtifact.getVersion())) {
                this.logger.debug("The dependency '{}' matches the ban exlusion filter", depArtifact);
                return false;
            }
        }
        
        if (this.failFast) {
            // plugin resolution downloads banned dependencies unless we fail now; not later
            throw new RuntimeException("Banned dependency detected: " + node + " => " + parents);
        } else {
            return true;
        }
    }
    
    private boolean matches(String exactFilter, String regexFilter, String value) {
        if (exactFilter == null && regexFilter == null) {
            return true;
        } else if (exactFilter != null) {
            return exactFilter.equals(value);
        } else {
            Pattern filterPattern = Pattern.compile(regexFilter);
            Matcher matcher = filterPattern.matcher(value);
            return matcher.matches();
        }
    }
    
    private boolean withinRange(VersionRange versionRange, String version) {
        boolean within = versionRange == null || versionRange.containsVersion(new DefaultArtifactVersion(version));
        this.logger.debug("Tested version range: {} <=> {}: {}", versionRange, version, within);
        return within;
    }
    
}
