/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.maven.ban;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Singleton;

import org.apache.maven.AbstractMavenLifecycleParticipant;
import org.apache.maven.MavenExecutionException;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.model.Plugin;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugin.PluginResolutionException;
import org.apache.maven.plugin.internal.PluginDependenciesResolver;
import org.apache.maven.plugins.annotations.Parameter;
import org.apache.maven.project.DefaultDependencyResolutionRequest;
import org.apache.maven.project.DependencyResolutionException;
import org.apache.maven.project.DependencyResolutionResult;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.ProjectDependenciesResolver;
import org.eclipse.aether.artifact.Artifact;
import org.eclipse.aether.artifact.DefaultArtifact;
import org.eclipse.aether.graph.Dependency;
import org.eclipse.aether.graph.DependencyFilter;
import org.eclipse.aether.graph.DependencyNode;
import org.eclipse.aether.impl.ArtifactResolver;
import org.eclipse.aether.impl.VersionRangeResolver;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.inteligr8.maven.ban.model.ImportConfig;

@Named("ban")
@Singleton
public class BanExtension extends AbstractMavenLifecycleParticipant {

    public static final String THIS_PLUGIN_GROUP_ID = "com.inteligr8";    
    public static final String THIS_PLUGIN_ARTIFACT_ID = "ban-maven-plugin";
    public static final String THIS_PLUGIN_KEY = THIS_PLUGIN_GROUP_ID + ":" + THIS_PLUGIN_ARTIFACT_ID;
    
    private Logger logger = LoggerFactory.getLogger(this.getClass());
    
    @Inject
    private ArtifactResolver artifactResolver;
    
    @Inject
    private VersionRangeResolver versionRangeResolver;
    
    @Inject
    private ProjectDependenciesResolver projDepResolver;
    
    @Inject
    private PluginDependenciesResolver pluginDepResolver;

    /**
     * The configuration is parsed manually.  This is here to prevent warning messages with IDEs and builders.
     */
    @Parameter(name = "import")
    private ImportConfig importConfig;

    @Parameter(name = "includes")
    private List<String> includes;

    @Parameter(name = "excludes")
    private List<String> excludes;
    
    @Override
    public void afterProjectsRead(MavenSession session) throws MavenExecutionException {
        BanConfiguration config = this.getConfiguration(session);
        if (config == null)
            return;
        
        BanDependencyFilter depFilter = new BanDependencyFilter(config.getIncludeArtifacts(), config.getExcludeArtifacts());
        depFilter.setFailFast(true);

        MavenProject project = session.getCurrentProject();
        
        try {
            for (Plugin plugin : project.getBuildPlugins()) {
                this.logger.debug("Evaluating plugin dependencies: {}", plugin);
                
                Artifact artifact = new DefaultArtifact(plugin.getId());
                DependencyNode depNodeRoot = this.pluginDepResolver.resolve(plugin, artifact, depFilter, project.getRemotePluginRepositories(), session.getRepositorySession());
                List<Dependency> bannedDependencies = this.crawlDependencyTree(depNodeRoot, depFilter);
                if (!bannedDependencies.isEmpty())
                    throw new MavenExecutionException("Banned dependencies were detected in plugin '" + plugin + "': " + bannedDependencies, project.getFile());
            }
        } catch (PluginResolutionException pre) {
            throw new MavenExecutionException(pre.getMessage(), pre);
        }

        depFilter.setFailFast(false);

        DefaultDependencyResolutionRequest request = new DefaultDependencyResolutionRequest(project, session.getRepositorySession());
        request.setResolutionFilter(depFilter);
        
        try {
            DependencyResolutionResult result = this.projDepResolver.resolve(request);
            List<Dependency> bannedDependencies = result.getDependencies();
            if (!bannedDependencies.isEmpty())
                throw new MavenExecutionException("Banned dependencies were detected: " + bannedDependencies, project.getFile());
        } catch (DependencyResolutionException dre) {
            throw new MavenExecutionException(dre.getMessage(), dre);
        }
    }
    
    private BanConfiguration getConfiguration(MavenSession session) throws MavenExecutionException {
        MavenProject project = session.getCurrentProject();
        Plugin plugin = project.getPlugin(THIS_PLUGIN_KEY);
        if (plugin == null)
            throw new MavenExecutionException("The plugin is executing but it cannot be found", project.getFile());
        
        if (!plugin.isExtensions()) {
            this.logger.warn("The '{}' plugin must be defined with '<extensions>true</extensions>'; ignoring plugin", plugin.getId());
            return null;
        } else {
            try {
                return new BanConfigurationParser(session, this.artifactResolver, this.versionRangeResolver, plugin);
            } catch (IOException | MojoFailureException e) {
                throw new MavenExecutionException(e.getMessage(), project.getFile());
            }
        }
    }
    
    private List<Dependency> crawlDependencyTree(DependencyNode depNode, DependencyFilter depFilter) {
        return this.crawlDependencyTree(depNode, new LinkedList<>(), depFilter);
    }
    
    private List<Dependency> crawlDependencyTree(DependencyNode depNode, List<DependencyNode> parentDepNodes, DependencyFilter depFilter) {
        List<Dependency> dependencies = new LinkedList<>();
        if (depFilter.accept(depNode, parentDepNodes))
            dependencies.add(depNode.getDependency());
        
        parentDepNodes.add(0, depNode);
        for (DependencyNode childDepNode : depNode.getChildren())
            dependencies.addAll(this.crawlDependencyTree(childDepNode, parentDepNodes, depFilter));
        parentDepNodes.remove(0);
        
        return dependencies;
    }
    
}
