/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.maven.ban;

import java.io.File;
import java.io.IOException;
import java.nio.file.DirectoryNotEmptyException;
import java.nio.file.FileVisitResult;
import java.nio.file.FileVisitor;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.function.Consumer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.inject.Inject;

import org.apache.maven.artifact.versioning.ArtifactVersion;
import org.apache.maven.artifact.versioning.DefaultArtifactVersion;
import org.apache.maven.artifact.versioning.VersionRange;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.model.Plugin;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.component.annotations.Component;
import org.eclipse.aether.impl.ArtifactResolver;
import org.eclipse.aether.impl.VersionRangeResolver;

import com.inteligr8.maven.model.ArtifactFilter;

@Mojo( name = "purge-repo", threadSafe = true )
@Component( role = org.apache.maven.plugin.Mojo.class )
public class PurgeRepoMojo extends AbstractMojo {
    
    @Inject
    private MavenSession session;
    
    @Inject
    private ArtifactResolver artifactResolver;
    
    @Inject
    private VersionRangeResolver versionRangeResolver;
    
    @Parameter(name = "skip", defaultValue = "false")
    private boolean skip = false;
    
    @Parameter(name = "dryRun", defaultValue = "false")
    private boolean dryRun = false;
    
    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        if (this.skip) {
            this.getLog().debug("Skipping purge of banned artifacts");
        } else {
            this.getLog().info("Purging banned artifacts from local repository: " + this.session.getLocalRepository().getBasedir());
            
            try {
                this.purge();
            } catch (IOException ie) {
                throw new MojoFailureException(ie.getMessage(), ie);
            }
        }
    }
    
    private void purge() throws MojoFailureException, IOException {
        List<Path> includePaths = new LinkedList<>();
        
        BanConfigurationParser config = this.getConfiguration();
        
        for (ArtifactFilter afilter : config.getIncludeArtifacts()) {
            if (afilter.getGroupId() == null) {
                this.getLog().warn("The purge does not support wildcard group ID specifications, so skipping it: " + afilter);
                continue;
            }
            
            Path groupPath = this.resolveGroupPath(afilter);
            if (groupPath == null)
                continue;
            
            List<Path> artifactPaths = this.resolveArtifactPaths(groupPath, afilter);
            for (Path artifactPath : artifactPaths)
                includePaths.addAll(this.resolveVersionPaths(artifactPath, afilter.getVersionRange()));
        }

        this.getLog().debug("May be purging all files in " + includePaths.size() + " paths");
        String regexDirectorySeparator = String.valueOf(File.separatorChar);
        if (File.separatorChar == '\\')
            regexDirectorySeparator += "\\";

        for (ArtifactFilter afilter : config.getExcludeArtifacts()) {
            StringBuilder regex = new StringBuilder();
            
            if (afilter.getGroupId() != null) {
                regex.append('^').append(this.resolveGroupPath(afilter));
            } else if (afilter.getGroupIdRegex() != null) {
                regex.append(afilter.getGroupIdRegex().replace("\\.", regexDirectorySeparator));
                if (regex.charAt(0) != '^')
                    regex.insert(0, '^');
            } else {
                regex.append("^.+");
            }
            
            regex.append(regexDirectorySeparator);

            if (afilter.getArtifactId() != null) {
                regex.append(afilter.getArtifactId());
            } else if (afilter.getArtifactIdRegex() != null) {
                regex.append(afilter.getArtifactIdRegex());
            } else {
                regex.append("[^").append(regexDirectorySeparator).append("]+");
            }
            
            Pattern pattern = Pattern.compile(regex.toString());
            
            Iterator<Path> i = includePaths.iterator();
            while (i.hasNext()) {
                Path path = i.next();
                
                Matcher matcher = pattern.matcher(path.toString());
                if (!matcher.find())
                    continue;
                
                // group/artifact match; now for version
                if (afilter.getVersionRange() == null) {
                    i.remove();
                } else {
                    String version = path.getFileName().toString();
                    if (afilter.getVersionRange().containsVersion(new DefaultArtifactVersion(version)))
                        i.remove();
                }
            }
        }

        Path repoPath = this.getRepositoryPath();
        if (this.dryRun) {
            this.getLog().info("DRYRUN: Would have deleted certain paths from local Maven cache: " + repoPath);
            this.getLog().info("DRYRUN: Would have deleted these paths: " + includePaths);
        } else {
            for (Path versionPath : includePaths) {
                Path fullVersionPath = repoPath.resolve(versionPath);
                if (Files.exists(fullVersionPath)) {
                    this.getLog().info("Deleting version from Maven cache: " + versionPath);
                    Files.walkFileTree(fullVersionPath, new DeleteNonMetadataVisitor());
                } else {
                    // this will probably never happen
                    this.getLog().debug("Maven cache does not exist: " + versionPath);
                }
            }
        }
	}
    
    private BanConfigurationParser getConfiguration() throws MojoFailureException, IOException {
        MavenProject project = this.session.getCurrentProject();
        Plugin plugin = project.getPlugin(BanExtension.THIS_PLUGIN_KEY);
        if (plugin == null)
            throw new MojoFailureException("The plugin is executing but it cannot be found");
        return new BanConfigurationParser(this.session, this.artifactResolver, this.versionRangeResolver, plugin);
    }
    
    private Path resolveGroupPath(ArtifactFilter afilter) {
        if (afilter.getGroupId() == null)
            return null;
        
        String[] pathElements = afilter.getGroupId().split("\\.");
        Path groupPath = Paths.get("");
        for (String pathElement : pathElements)
            groupPath = groupPath.resolve(pathElement);
        
        Path fullGroupPath = this.getRepositoryPath().resolve(groupPath);
        if (!Files.exists(fullGroupPath)) {
            this.getLog().debug("The group path does not exist, so nothing to purge: " + fullGroupPath);
            return null;
        }
        
        return groupPath;
    }
    
    private List<Path> resolveArtifactPaths(Path groupPath, ArtifactFilter afilter) throws IOException {
        Path repoPath = this.getRepositoryPath();
        
        if (afilter.getArtifactId() != null) {
            Path artifactPath = groupPath.resolve(afilter.getArtifactId());
            Path fullArtifactPath = repoPath.resolve(artifactPath);
            if (Files.exists(fullArtifactPath)) {
                return Arrays.asList(artifactPath);
            } else {
                this.getLog().debug("The artifact path does not exist, so nothing to purge: " + fullArtifactPath);
                return Collections.emptyList();
            }
        }
        
        Pattern artifactPattern = afilter.getArtifactIdRegex() == null ? null : Pattern.compile(afilter.getArtifactIdRegex());
        List<Path> paths = new LinkedList<>();
        
        if (artifactPattern == null)
            this.getLog().debug("All artifact directories in '" + groupPath + "' qualify as included");
        
        Files.list(repoPath.resolve(groupPath)).forEach(new Consumer<Path>() {
            @Override
            public void accept(Path fullArtifactPath) {
                if (artifactPattern == null) {
                    // these may include sub-groups and not just artifacts
                    // which will lead to paths with artifacts as versions
                    // so we are looping through versions to see if it is indeed an artifact
                    try {
                        Files.list(fullArtifactPath).findFirst().ifPresent(new Consumer<Path>() {
                            @Override
                            public void accept(Path fullVersionPath) {
                                if (Files.exists(fullVersionPath.resolve("_remote.repositories")))
                                    paths.add(repoPath.relativize(fullArtifactPath));
                            }
                        });
                    } catch (IOException ie) {
                        getLog().error(ie.getMessage(), ie);
                    }
                } else {
                    Matcher matcher = artifactPattern.matcher(fullArtifactPath.getFileName().toString());
                    if (matcher.matches()) {
                        getLog().debug("The artifact directory '" + fullArtifactPath.getFileName() + "' qualifies as included");
                        paths.add(repoPath.relativize(fullArtifactPath));
                    }
                }
            }
        });
        
        return paths;
    }
    
    private List<Path> resolveVersionPaths(Path artifactPath, VersionRange versionRange) throws IOException {
        Path repoPath = this.getRepositoryPath();
        List<Path> paths = new LinkedList<>();
        
        if (versionRange == null)
            this.getLog().debug("All artifact version directories in '" + artifactPath + "' qualify as included");
        
        Files.list(repoPath.resolve(artifactPath)).forEach(new Consumer<Path>() {
            @Override
            public void accept(Path t) {
                if (versionRange == null) {
                    paths.add(repoPath.relativize(t));
                } else {
                    ArtifactVersion artifactVersion = new DefaultArtifactVersion(t.getFileName().toString());
                    if (versionRange.containsVersion(artifactVersion)) {
                        getLog().debug("The artifact version directory '" + t.getFileName() + "' qualifies as included");
                        paths.add(repoPath.relativize(t));
                    }
                }
            }
        });
        
        return paths;
    }
    
    private Path getRepositoryPath() {
        return new File(this.session.getLocalRepository().getBasedir()).toPath();
    }
    
    
    
    private class DeleteNonMetadataVisitor implements FileVisitor<Path> {
        
        private final Pattern versionPathPattern = Pattern.compile("/([^/]+)/([^/]+)/[^/]+$");
        
        @Override
        public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) throws IOException {
            return FileVisitResult.CONTINUE;
        }
        
        @Override
        public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
            if (attrs.isDirectory()) {
                getLog().warn("An unexpected directory was found: " + file);
                return FileVisitResult.SKIP_SUBTREE;
            }

            Matcher matcher = this.versionPathPattern.matcher(file.toString());
            if (!matcher.find()) {
                getLog().debug("Ignoring file from purge: " + file);
                return FileVisitResult.CONTINUE;
            }
            
            String artifactId = matcher.group(1);
            String version = matcher.group(2);
            String includeName = artifactId + "-" + version;
            String excludeName = artifactId + "-" + version + ".pom";
            getLog().debug("artifact-version: " + includeName);
            
            if (file.getFileName().toString().startsWith(includeName) &&
                    !file.getFileName().toString().startsWith(excludeName)) {
                try {
                    getLog().info("Deleting artifact: " + file);
                    Files.delete(file);
                } catch (IOException ie) {
                    getLog().debug(ie);
                    getLog().warn("The file failed to delete: " + file);
                    return FileVisitResult.SKIP_SIBLINGS;
                }
            }
            
            return FileVisitResult.CONTINUE;
        }
        
        @Override
        public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
            return FileVisitResult.CONTINUE;
        }
        
        @Override
        public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
            try {
                Files.delete(dir);
            } catch (DirectoryNotEmptyException dnee) {
                getLog().debug("The folder will not be deleted as it is not empty: " + dir);
            } catch (IOException ie) {
                getLog().debug(ie);
                getLog().warn("The folder failed to delete: " + dir);
            }
            
            return FileVisitResult.CONTINUE;
        }
        
    }
    
}
